%% ERGO COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1993 by The University of Queensland, Queensland 4072 Australia
%% 
%% 
%% Permission to use, copy and distribute this software for any
%% non-commercial purpose and without fee is hereby granted, provided
%% that the above copyright notice and this permission notice and
%% warranty disclaimer appear in all copies and in supporting
%% documentation, and that the name of the University of Queensland not
%% be used in advertising or publicity pertaining to distribution of
%% the software without specific, written prior permission.
%% 
%% Source code modifications are prohibited except where written 
%% agreement has been given in advance by the University of Queensland.
%% 
%% The University of Queensland disclaims all warranties with regard to
%% this software, including all implied warranties of merchantability
%% and fitness.  In no event shall the University of Queensland be
%% liable for any special, indirect or consequential damages or any
%% damages whatsoever resulting from loss of use, data or profits,
%% whether in an action of contract, negligence or other tortious
%% action, arising out of or in connection with the use or performance
%% of this software.
%% 
%% CVS show.ql,v 1.10 1993/12/10 02:19:08 marku Exp


%% theory_topic(?Topic, -Arity_or_Global).
%% global_topic(?Topic, -Arity).
%%
%%   These two predicates can be used to check whether a given
%%   topic is valid, or to backtrack through all possible topics.
%%
%%   Arity is an integer, giving the arity of Topic in global.erg.
%%
%%   Arity_or_Global is either an integer, which is the arity of
%%   Topic in each theory file, or a term of the form global/N,
%%   which means that Topic is actually stored in global.erg and
%%   has arity N.

theory_topic(ancestor, global/2).
theory_topic(parent,   global/2).
theory_topic(symbol,   global/4).
theory_topic(T, N) :-
	theory_fact(T / N).

global_topic(theory, 2).
global_topic(nfi, 0).
global_topic(T, N) :-
	global_fact(T / N),
	\+ (theory_topic(T, _)).


%% show_topic(+Head, +Body, +Indent).
%%
%%   This predicate determines how each clause of information is displayed.
%%   The last case can handle everything, but the earlier cases are there
%%   to pretty print tactics, heuristics etc.

show_topic(axiom(Name,Stmt), Body, Indent) :-
	!,
	writelist(Indent),
	writelist(['axiom ', Name, ' === ', Stmt]),
	show_topic_body(Body, Indent).

show_topic(theorem(Name,Stmt), Body, Indent) :-
	!,
	writelist(Indent),
	writelist(['theorem ', Name, ' === ', Stmt]),
	show_topic_body(Body, Indent).


show_topic(envir(Name, Cutoff, Levels), true, Indent) :-
	!,
	writelist(Indent),
	writelist_indent(['envir(',Name,', ',Cutoff,','], Indent),nl,
	write_envir_levels(Levels, 1, ['    ' | Indent]),
	writelist(Indent),
	writelist(').'),nl.

show_topic(tactic(Name, Doc, Func, _, _, _, _), Body, Indent) :-
	!,
	functor(Func, _, Arity),
	writelist(Indent),
	writelist_indent(['tactic ', Name, '/', Arity, nl], Indent),
	writelist_indent(['    ' | Doc], ['    ' | Indent]),
	show_topic_body(Body, Indent).

show_topic(heuristic(Call, DefLevel, Doc, _, _, _, _), Body, Indent) :-
	!,
	Doc \== sub_heuristic,
	writelist(Indent),
	writelist_indent(['heuristic ',Call,
			'  Default level = ',DefLevel,nl], Indent),
	writelist_indent(['    ' | Doc], ['    '|Indent]),
	show_topic_body(Body, Indent).

show_topic(Head, Body, Indent) :-
	writelist(Indent),
	writelist(Head),
	show_topic_body(Body, Indent).


%% show_topic_body(+Body, +Indent).
%%
show_topic_body(true, _) :-
	!,
	write('.'),
	nl.

show_topic_body(Body, Indent) :-
	writelist_indent([nl, '    :- ', Body, '.'], Indent),
	nl.



%% show.
%%
%%   lists all the available (global and theory specific) topics.

show :- 
	writeln('Usage: show(Topic)'),
	writeln('       show(Topic, Theories)'),
	nl,
	write('Available global topics are:'),
	global_topic(Topic, _),
	writelist([' ', Topic]),
	fail.

show :-
	write('\n\nTheory specific topics are:'),
	theory_topic(Topic, _),
	writelist([' ', Topic]),
	fail.

show :- nl.


show(Topics) :-
        \+ (elem(T, Topics),
	    (theory_topic(T, _) ->
		check_or_writelist(current_dir(CD),
			['Error: no current theory',nl])
	    ;
	        true
	    ),
	    do_topic(T, CD),
	    fail).

show(Topics, all) :-
	!,
	check_or_writelist(current_dir(CD),
		['Error: no current theory',nl]),
	(ancestor : [CD, Parents]),
	show(Topics, [CD|Parents]).

show(Topics, Theories) :-
	\+ (elem(Thy, Theories),
	    writelist([nl,'Theory: ', Thy, nl]),
	    check_or_writelist((ancestor : [Theory,_]),
	        ['Error: unknown theory:', T, nl]),
	    elem(T, Topics),
	    check_or_writelist(theory_topic(T, _),
	        ['Error: ', T, ' is not a theory specific topic',nl]),
	    do_topic(T, Thy),
	    fail).


%% do_topic(+Topic, ?Theory).
%%
%%   This writes out global or theory specific information about Topic.
%%   If Topic is theory specific, then Theory must be instantiated to
%%   the name of a theory.

do_topic(Topic, Theory) :-
	theory_topic(Topic, global/Arity),
	!,
        length([Theory|Args], Arity),
	clause(Topic : [Theory|Args], Body),
	Clause =.. [Topic|Args],
	show_topic(Clause, Body, '    ').

do_topic(Topic, Theory) :-
	theory_topic(Topic, Arity),
	integer(Arity),
	!,
	check_or_writelist(([status,Theory] : _),
	    ['Warning: theory', Theory, ' is not loaded yet.',nl]),
        length(Args, Arity),
	find_at(Topic, Theory, Args, Body),
	Clause =.. [Topic|Args],
	show_topic(Clause, Body, '    ').

do_topic(theory, _) :-
	!,
	(ancestor : [Thy, Ancestors]),
	show_topic(theory(Thy, Ancestors), true, '    ').

do_topic(nfi, _) :-
	!,
	nfi.

do_topic(Topic, _) :-
	global_topic(Topic, Arity),
	!,
        length(Args, Arity),
	clause(Topic : Args, Body),
	Clause =.. [Topic|Args],
	show_topic(Clause, Body, '    ').

do_topic(Topic, _) :-
	writelist(['Unknown topic: ', Topic, nl]),
	fail.
