%% ERGO COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1993 by The University of Queensland, Queensland 4072 Australia
%% 
%% 
%% Permission to use, copy and distribute this software for any
%% non-commercial purpose and without fee is hereby granted, provided
%% that the above copyright notice and this permission notice and
%% warranty disclaimer appear in all copies and in supporting
%% documentation, and that the name of the University of Queensland not
%% be used in advertising or publicity pertaining to distribution of
%% the software without specific, written prior permission.
%% 
%% Source code modifications are prohibited except where written 
%% agreement has been given in advance by the University of Queensland.
%% 
%% The University of Queensland disclaims all warranties with regard to
%% this software, including all implied warranties of merchantability
%% and fitness.  In no event shall the University of Queensland be
%% liable for any special, indirect or consequential damages or any
%% damages whatsoever resulting from loss of use, data or profits,
%% whether in an action of contract, negligence or other tortious
%% action, arising out of or in connection with the use or performance
%% of this software.
%% 
%% CVS translate.ql,v 1.11 1993/12/10 02:19:14 marku Exp


/*
 * Translation of documented Dylan commands into internal commands
 * ===============================================================
 *
 * NB.	1. Within function and constant declarations and definitions,
 *         names of the form 'abc,xyz' are split at the comma, with
 *	   the first part (eg. 'abc') being used for the function or operator
 *	   being defined and the second part (eg. 'xyz') being used when
 *	   axiom names are constructed.  For example, a name such as
 *	   '+,add' might be used to define addition.
 *
 *	   This 'feature' will probably disappear soon...
 *
 *	2. All expressions (eg. types and the bodies of definitions)
 *         undergo expression translation upon input.
 *
 *	   An abbreviation expansion is performed on every
 *	   subexpression.  That is, any subexpression E is translated
 *	   to E' if there is a declaration of the form:
 *
 *		abbreviation E === E'.
 *
 *	   WARNING: these abbreviations are expanded repeatedly,
 *	   so you have to be careful not to define a series of
 *	   abbreviations that can cause a loop!
 *
 * BUG:   This implementation can generate several Dylan commands
 *         D1,D2,... from a single input command.  If some of those 
 *         commands (say D1..Dk) succeed, but D(k+1) fails, then 
 *         the original input command will fail, but will not undo
 *         the effects of D1..Dk.  This may mean that the original
 *         input command cannot legally be re-entered!
 */


?- object_var('x').
?- op(900, quant, is_local).

?- op(1200,xfy,===).

?- op(999,fx,abbreviation).
?- op(999,fx,constant).
?- op(999,fx,absolute_constant).
?- op(999,fx,function).
?- op(999,fx,absolute_function).
?- op(999,fx,axiom).
?- op(999,fx,postulate).
?- op(999,fx,theorem).

?- op(998,xfx,leftassoc).
?- op(998,xfx,rightassoc).
?- op(998,xfx,nonassoc).
?- op(998,xfx,prefix).
?- op(998,xfx,postfix).

:- declare((===)/2).
:- declare(reload_translate/1).
:- declare((constant)/1).
:- declare((absolute_constant)/1).
:- declare((function)/1).
:- declare((absolute_function)/1).
:- declare((quantifier)/1).
:- declare((quantifier)/2).


/*
 * DEFINITIONS
 * ===========
 */

(abbreviation E === E2) :-
        tactic_assert(abbreviation,[E,E2]).

(constant C === Body) :-
	parse_head(C, Name, _, []),
	!,
	expr_de_abbreviate(Body,Body2),
	make_objvars_disjoint( Body2 ),
	deff(Name, [], Body2, default, 0).

(function F === Body) :-
	parse_function(F, Call, Prec, Opinfo, Arity),
	parse_head(Call, Name, _, Args),
	length(Args,Arity),
	!,
	objvarlist(Args),   % also checks they are not typed.
	expr_de_abbreviate(Body,Body2),
	make_objvars_disjoint( [Args,Body2] ),
	deff(Name, Args, Body2, Prec, Opinfo).

(quantifier(Q,x,B) === Body) :-
	expr_de_abbreviate(Body,Body2),
	make_objvars_disjoint( [x,Body2] ),
	defq(Q,x,B,Body2,900).

(quantifier(Q,x,B,Prec) === Body) :-
	expr_de_abbreviate(Body,Body2),
	make_objvars_disjoint( [x,Body2] ),
	defq(Q,x,B,Body2,Prec).


/*
 * AXIOMS AND THEOREMS
 * ===================
 */

(axiom Name === (Stmt :- SideConds)) :-
	!,
	once(expr_de_abbreviate(Stmt,Stmt2)),
	%% WARNING: we should really use a more specialised 
	%%          tactic/code de-abbreviator here.
	once(expr_de_abbreviate(SideConds,SideConds2)),
	once(make_objvars_disjoint((Stmt2 :- SideConds2))),
	mkaxiom(Name,Stmt2,SideConds2).

(axiom Name === Stmt) :-
	%% not Stmt = (_ :- _)
	!,
	once(expr_de_abbreviate(Stmt,Stmt2)),
	once(make_objvars_disjoint(Stmt2)),
	mkaxiom(Name,Stmt2).

/* A synonym for `axiom' */
(postulate Name === S) :-
	(axiom Name === S).


(theorem Name === (Stmt :- Tactic)) :-
	!,
	once(expr_de_abbreviate(Stmt,Stmt2)),
	%% WARNING: we should really use a more specialised 
	%%          tactic/code de-abbreviator here.
	once(expr_de_abbreviate(Tactic,Tactic2)),
	once(make_objvars_disjoint((Stmt2 :- Tactic2))),
	theorem(Name,Stmt2,Tactic2).

(theorem Name === Stmt) :-
	%% not Stmt = (_ :- _)
	!,
	once(expr_de_abbreviate(Stmt,Stmt2)),
	once(make_objvars_disjoint(Stmt2)),
	theorem(Name,Stmt2).


/*
 * DECLARATIONS
 * ============
 *
 *  NB. object_var/1, equivalence/1, stronger_than/2 and openrule/6
 *  are all handled directly by Dylan.
 */

quantifier(Q) :-
	quantifier(Q,900).

quantifier(Q,Prec) :-
	atom(Q),
	primq(Q,Prec).


constant(Atom:Type) :-
	atom(Atom),
	!,
	AxStmt = (Name:Type),
	parse_head(Atom, Name, AxName, []),
	const(Name),
	once(expr_de_abbreviate(AxStmt,AxStmt2)),
	once(make_objvars_disjoint(AxStmt2)),
	makeaxiom(AxName,"_type", AxStmt2).

constant(Atom) :-
	atom(Atom),
	!,
	parse_head(Atom, Name, _, []),
	const(Name).

absolute_constant(Atom:Type) :-
	constant(Atom:Type),
	parse_head(Atom, Name, AxName, []),
	makeaxiom(AxName,"_abs", is_abs(Name)).

absolute_constant(Atom) :-
	atom(Atom),
	!,
	parse_head(Atom, Name, AxName, []),
	const(Name),
	makeaxiom(AxName,"_abs", is_abs(Name)).


function(F) :-
	parse_function(F, Call, Prec, Opinfo, Arity),
	function2(Call, Prec, Opinfo, _Name, _AxNameStr, Arity).

absolute_function(F) :-
	parse_function(F, Call, Prec, Opinfo, Arity),
	function2(Call, Prec, Opinfo, Name, AxNameStr, Arity),

	/* generate axiom: (T;Name(A1,A2,...)) = Name((T;A1), (T;A2),...) */
         length(L1,Arity),
         FuncCall1 =.. [Name|L1],
         shift_args(L1,T,T_L1),
         FuncCall1_T =.. [Name|T_L1],
       	 !,
        (inherited_symbol(';', 2, _) -> 
	         makeaxiom(AxNameStr,"_local", (T ; FuncCall1) = FuncCall1_T)
	       ; true),

	/* generate axiom: (is_local x Arg1) and ...
         *                 => (is_local x Name(Arg1,...)
         */
	make_is_local_list(L1,x,ArgsLocal),
	!,
	makeaxiom(AxNameStr,"_is_local", ArgsLocal => (is_local x FuncCall1)).



/******************** AUXILLARY PREDICATES ***********************/

/*
 * function2(Declaration, Prec, Opinfo, Name, AxiomNameString, Arity).
 *
 *	Declaration is a (possibly typed) function call.  eg. f(A,B):Type.
 *
 *      Prec may be an integer precedence (then Opinfo must be either fy,
 *      yf, xfx, yfx or xfy) or the atom 'default' (then Opinfo must be
 *      the arity of the function -- eg. by binding it to Arity).
 *
 *      Name, AxiomNameString and Arity are outputs, but can be bound
 *      upon input as a cross check if desired.
 *
 *      Name is the name of the function, returned as an atom.
 *
 *      AxiomNameString is an alternative name for the function, returned
 *      as a string.  It should be used within axiom names.
 *
 *      Arity is the number of arguments in Declaration.
 */

function2(Head:Type, Prec, Opinfo, Name, AxNameStr, Arity) :-
	!,
	parse_head(Head, Name, AxNameStr, Args),
	function2(Head, Prec, Opinfo, Name, AxNameStr, Arity),
	/*
	 * now generate the type axiom. 
	 * Eg. Head=f(A:T):T2 generates: 
	 *	mkaxiom(f_type,A:T => f(A):T2).
	 */
	AxStmt = (ArgPred => (FuncCall:Type)),
	make_argpred(Args,ArgPred),
	split_args(Args,ArgNames,_),
	FuncCall =.. [Name|ArgNames],
	once(expr_de_abbreviate(AxStmt, AxStmt2)),
	once(make_objvars_disjoint(AxStmt2)),
	makeaxiom(AxNameStr,"_type", AxStmt2).

function2(Head, Prec, Opinfo, Name, AxNameStr, Arity) :-
	/* not Head = Hd:Type */
	parse_head(Head, Name, AxNameStr, Args),
	length(Args,Arity),
	primf(Name,Prec,Opinfo).




/*
 * shift_args(Arglist, T, Shifted_Arglist)
 *
 *	Eg. shift_args([A,B], T, [(T;A), (T;B)]).
 */

shift_args([],_,[]).

shift_args([Arg|Args], T, [(T;Arg)|Pred]) :-
	shift_args(Args, T, Pred).



/*
 * make_is_local_list(+Varlist, +ObjVar, -Pred).
 *
 *	Eg. make_is_local_list([A,B], x, (is_local x A) and (is_local x B)).
 *	Requires: Varlist must have length >= 1.
 */

make_is_local_list([A], x, (is_local x A)) :-
	!.

make_is_local_list([A|AS], x, (is_local x A) and Pred) :-
	/* not AS=[] */
	make_is_local_list(AS, x, Pred).



/*
 * split_args(TypedVarList, VarList, TypeList).
 *
 *	Eg. split_args([A:T1,B:T2], [A,B], [T1,T2]).
 */

split_args([],[],[]).

split_args([(Arg:Type)|NTs], [Arg|Ns], [Type|Ts]) :-
	split_args(NTs, Ns, Ts).



/*
 * make_argpred(TypedVarList, Pred).
 *
 *      Turns TypedVarList into a conjunction of type predicates.
 *
 *	Eg. make_argpred([a:A,b:B], a:A and b:B).
 *
 *	Requires: TypedVarlist must have length >= 1.
 */

make_argpred([Name:Type], Name:Type).

make_argpred([(Name:Type)|Args], (Name:Type) and ArgsPred) :-
	make_argpred(Args, ArgsPred).



/* 
 * parse_function(Decl, Call, Prec, Opinfo, Arity)
 *
 * Strips the precedence and operator information off Decl and
 * returns it in Prec and Opinfo, respectively, and returns the
 * remaining function call in Call.   Arity is instantiated to 1 or 2
 * for operators, and is bound to Opinfo for non-operators.
 */

parse_function(Prec leftassoc Call,  Call, Prec, yfx, 2) :- !.
parse_function(Prec rightassoc Call, Call, Prec, xfy, 2) :- !.
parse_function(Prec nonassoc Call,   Call, Prec, xfx, 2) :- !.
parse_function(Prec prefix Call,     Call, Prec, fy,  1) :- !.
parse_function(Prec postfix Call,    Call, Prec, yf,  1) :- !.
parse_function(Call,                 Call, default, Arity, Arity).


/*
 * parse_head(Head, Name, AxNameStr, Args).
 *
 *	This is roughly equivalent to Head =.. [Name|Args].
 *	However, it also converts Name to a string and passes
 *	it out as AxNameStr.  In the special case where the
 *	functor of Head is an atom containing a comma (e.g. '<,lt')
 *	that functor is split into two parts so that Name = '<' and
 *	AxNameStr = "lt".
 */

parse_head(Head, Name, AxNameStr, []) :-
	atom(Head),
	!,
	translate_name(Head, Name, AxNameStr).

parse_head(Head, Name, AxNameStr, Args) :-
	Head =.. [F|Args],
	translate_name(F, Name, AxNameStr).

/*
 * translate_name/3 splits Atom into the two comma-separated parts, 
 * or leaves it unchanged otherwise.
 */
translate_name(Atom, Name, AxNameStr) :-
	atom_chars(Atom,Str),
	append(NameStr, [44|AxNameStr], Str),
	\+ (NameStr = []),
	\+ (AxNameStr = []),
	!,
	atom_chars(Name, NameStr).

translate_name(Atom, Atom, AxNameStr) :-
	atom_chars(Atom,AxNameStr).

	
/*
 * objvarlist(VarList).
 *
 *	Instantiate a list of distinct Prolog variables into a list of
 *	distinct object variables (by asserting a not_free_in constraint
 *	between each pair of object variables).
 *
 *	The implementation assumes that `x' is an object variable.
 */

objvarlist([]).

objvarlist([x|Vs]) :-
	objvarlist(Vs),
	x not_free_in Vs.


/*
 * makeaxiom(Name,Suffix,Pred).
 *
 *	An interface to mkaxiom that constructs the axiom name
 *	from Name and Suffix.
 */	

makeaxiom(Name,Suffix,Pred) :-
	atom(Name),
	!,
	write('Internal translator error with: '),
	write(makeaxiom(Name,Suffix,Pred)).

makeaxiom(Name,Suffix,Pred) :-
	append(Name,Suffix, AxNameStr),
	atom_chars(AxName, AxNameStr),
	!,
	mkaxiom(AxName,Pred).


make_objvars_disjoint(Term) :-
	'$collect_variables'(Term, object_var, [], Ovars),
	objvarlist(Ovars).

